@@ -69,6 +69,8 @@ from ..models import Spotlight
 from ..models import User
 from ..models import UserQueryType
 from ..models import WikiPage
+from .repository import BaseTokenRepository
+from .repository import SimpleTokenRepository
 
 if TYPE_CHECKING:
     from types import TracebackType
@@ -83,6 +85,27 @@ __all__ = ("Client",)
 ClientRequestType = Literal["GET", "POST", "DELETE", "PUT", "PATCH"]
 
 
+def get_content_type(content_type: str) -> str:
+    """Returns the content type."""
+    return content_type.split(";")[0]
+
+
+def prepare_client(func: Callable) -> Callable:
+    """A decorator that prepares the client, to be used as:
+    @prepare_client
+    """
+
+    @functools.wraps(func)
+    async def _prepare_client(self: Client, *args: Any, **kwargs: Any) -> Any:
+        if self._session is None:
+            self._session = aiohttp.ClientSession()
+        if not await self._token_exists():
+            await self._add_token(self._initial_token)
+        return await func(self, *args, **kwargs)
+
+    return _prepare_client
+
+
 def check_token(func: Callable) -> Callable:
     """
     A decorator that checks the current token, to be used as:
@@ -91,7 +114,8 @@ def check_token(func: Callable) -> Callable:
 
     @functools.wraps(func)
     async def _check_token(self: Client, *args: Any, **kwargs: Any) -> Any:
-        if datetime.utcnow() > self.token.expires_on:
+        token = await self.get_current_token()
+        if datetime.utcnow() > token.expires_on:
             await self._refresh()
         return await func(self, *args, **kwargs)
 
@@ -107,10 +131,11 @@ def requires_scope(required_scopes: Scopes, any_scope: bool = False) -> Callable
     def _requires_scope(func: Callable) -> Callable:
         @functools.wraps(func)
         async def _wrap(self: Client, *args: Any, **kwargs: Any) -> Any:
+            token = await self.get_current_token()
             if any_scope:
-                if not (required_scopes & self.token.scopes):
+                if not (required_scopes & token.scopes):
                     raise APIException(403, "Missing required scopes.")
-            elif required_scopes & self.token.scopes != required_scopes:
+            elif required_scopes & token.scopes != required_scopes:
                 raise APIException(403, "Missing required scopes.")
 
             return await func(self, *args, **kwargs)
@@ -127,28 +152,36 @@ class Client(Eventable):
         See below
 
     :Keyword Arguments:
+        * *token_repository* (``aiosu.v2.repository.BaseTokenRepository``) --
+            Optional, defaults to ``aiosu.v2.repository.SimpleTokenRepository()``
+        * *session_id* (``int``) --
+            Optional, ID of the session to search in the repository, defaults to 0
         * *client_id* (``int``) --
             Optional, required for client credentials
         * *client_secret* (``str``) --
             Optional, required for client credentials
         * *base_url* (``str``) --
             Optional, base API URL, defaults to "https://osu.ppy.sh"
-        * *scopes* (``aiosu.models.Scopes``) --
-            Optional, defaults to ``Scopes.PUBLIC | Scopes.IDENTIFY``. Ignored if token is provided.
         * *token* (``aiosu.models.oauthtoken.OAuthToken``) --
             Optional, defaults to client credentials if not provided
         * *limiter* (``tuple[int, int]``) --
             Optional, rate limit, defaults to (600, 60) (600 requests per minute)
     """
 
-    def __init__(self, **kwargs: Any) -> None:
+    def __init__(
+        self,
+        **kwargs: Any,
+    ) -> None:
         super().__init__()
         self._register_event(ClientUpdateEvent)
+        self._token_repository: BaseTokenRepository = kwargs.pop(
+            "token_repository",
+            SimpleTokenRepository(),
+        )
+        self.session_id: int = kwargs.pop("session_id", 0)
         self.client_id: int = kwargs.pop("client_id", None)
         self.client_secret: str = kwargs.pop("client_secret", None)
-        unsafe_scopes: Scopes = kwargs.pop("scopes", Scopes.PUBLIC | Scopes.IDENTIFY)
-        self.token: OAuthToken = kwargs.pop("token", OAuthToken(scopes=unsafe_scopes))
-        self.scopes = self.token.scopes
+        self._initial_token: OAuthToken = kwargs.pop("token", OAuthToken())
         self.base_url: str = kwargs.pop("base_url", "https://osu.ppy.sh")
         max_rate, time_period = kwargs.pop("limiter", (600, 60))
         if (max_rate / time_period) > (1000 / 60):
@@ -188,37 +221,50 @@ class Client(Eventable):
 
         return _on_client_update
 
-    def _get_headers(self) -> dict[str, str]:
+    async def get_current_token(self) -> OAuthToken:
+        """Get the current token"""
+        return await self._token_repository.get(self.session_id)
+
+    async def _add_token(self, token: OAuthToken) -> None:
+        """Add a token to the current session"""
+        await self._token_repository.add(self.session_id, token)
+
+    async def _update_token(self, token: OAuthToken) -> None:
+        """Update the current token"""
+        await self._token_repository.update(self.session_id, token)
+
+    async def _token_exists(self) -> bool:
+        """Check if a token exists for the current session"""
+        return await self._token_repository.exists(self.session_id)
+
+    async def _get_headers(self) -> dict[str, str]:
+        token = await self.get_current_token()
         return {
-            "Authorization": f"Bearer {self.token.access_token}",
+            "Authorization": f"Bearer {token.access_token}",
             "Content-Type": "application/json",
             "Accept": "application/json",
         }
 
+    async def _refresh_auth_data(self) -> dict[str, Union[str, int]]:
+        token = await self.get_current_token()
+        return {
+            "client_id": self.client_id,
+            "client_secret": self.client_secret,
+            "grant_type": "refresh_token",
+            "refresh_token": token.refresh_token,
+        }
+
     def _refresh_guest_data(self) -> dict[str, Union[str, int]]:
         return {
             "client_id": self.client_id,
             "client_secret": self.client_secret,
             "grant_type": "client_credentials",
-            "scope": str(self.scopes),
-        }
-
-    def _refresh_auth_data(self) -> dict[str, Union[str, int]]:
-        return {
-            "client_id": self.client_id,
-            "client_secret": self.client_secret,
-            "grant_type": "refresh_token",
-            "refresh_token": self.token.refresh_token,
+            "scope": "public",
         }
 
     async def _request(
         self, request_type: ClientRequestType, *args: Any, **kwargs: Any
     ) -> Any:
-        if self._session is None:
-            self._session = aiohttp.ClientSession(
-                headers=self._get_headers(),
-            )
-
         req: dict[str, Callable] = {
             "GET": self._session.get,
             "POST": self._session.post,
@@ -230,7 +276,7 @@ class Client(Eventable):
         async with self._limiter:
             async with req[request_type](*args, **kwargs) as resp:
                 body = await resp.read()
-                content_type = resp.headers.get("content-type", "")
+                content_type = get_content_type(resp.headers.get("content-type", ""))
                 if resp.status != 200:
                     json = orjson.loads(body)
                     raise APIException(resp.status, json.get("error", ""))
@@ -240,20 +286,20 @@ class Client(Eventable):
                     return BytesIO(body)
                 if content_type == "text/plain":
                     return body.decode()
-                raise APIException(415, "Unhandled Content Type")
+                raise APIException(415, f"Unhandled Content Type '{content_type}'")
 
     async def _refresh(self) -> None:
         r"""INTERNAL: Refreshes the client's token
 
         :raises APIException: Contains status code and error message
         """
-        old_token = self.token
+        old_token = await self.get_current_token()
         url = f"{self.base_url}/oauth/token"
         headers = {"Content-Type": "application/x-www-form-urlencoded"}
 
         data = {}
-        if self.token.can_refresh:
-            data = self._refresh_auth_data()
+        if old_token.can_refresh:
+            data = await self._refresh_auth_data()
         else:
             data = self._refresh_guest_data()
 
@@ -262,26 +308,32 @@ class Client(Eventable):
                 async with temp_session.post(url, data=data) as resp:
                     try:
                         body = await resp.read()
-                        content_type = resp.headers.get("content-type", "")
+                        content_type = get_content_type(
+                            resp.headers.get("content-type", ""),
+                        )
                         if content_type != "application/json":
-                            raise APIException(415, "Unhandled Content Type")
+                            raise APIException(
+                                415,
+                                f"Unhandled Content Type '{content_type}'",
+                            )
                         json = orjson.loads(body)
                         if resp.status != 200:
                             raise APIException(resp.status, json.get("error", ""))
                         if self._session:
                             await self._session.close()
-                        self.token = OAuthToken.parse_obj(json)
-                        self.token.scopes = self.scopes
+                        new_token = OAuthToken.parse_obj(json)
+                        await self._update_token(new_token)
                         self._session = aiohttp.ClientSession(
-                            headers=self._get_headers(),
+                            headers=await self._get_headers(),
                         )
                     except aiohttp.client_exceptions.ContentTypeError:
                         raise APIException(403, "Invalid token specified.")
 
         await self._process_event(
-            ClientUpdateEvent(client=self, old_token=old_token, new_token=self.token),
+            ClientUpdateEvent(client=self, old_token=old_token, new_token=new_token),
         )
 
+    @prepare_client
     async def get_featured_artists(self, **kwargs: Any) -> ArtistResponse:
         r"""Gets the current featured artists.
 
@@ -331,6 +383,7 @@ class Client(Eventable):
             resp.next = partial(self.get_featured_artists, **kwargs)
         return resp
 
+    @prepare_client
     async def get_seasonal_backgrounds(self) -> SeasonalBackgroundSet:
         r"""Gets the current seasonal background set.
 
@@ -342,6 +395,7 @@ class Client(Eventable):
         json = await self._request("GET", url)
         return SeasonalBackgroundSet.parse_obj(json)
 
+    @prepare_client
     async def get_changelog_listing(self, **kwargs: Any) -> ChangelogListing:
         r"""Gets the changelog listing.
 
@@ -382,6 +436,7 @@ class Client(Eventable):
             resp.next = partial(self.get_changelog_listing, **kwargs)
         return resp
 
+    @prepare_client
     async def get_changelog_build(self, stream: str, build: str) -> Build:
         r"""Gets a specific build from the changelog.
 
@@ -395,6 +450,7 @@ class Client(Eventable):
         json = await self._request("GET", url)
         return Build.parse_obj(json)
 
+    @prepare_client
     async def lookup_changelog_build(
         self, changelog_query: Union[str, int], **kwargs: Any
     ) -> Build:
@@ -424,6 +480,7 @@ class Client(Eventable):
         json = await self._request("GET", url, params=params)
         return Build.parse_obj(json)
 
+    @prepare_client
     async def get_news_listing(self, **kwargs: Any) -> NewsListing:
         r"""Gets the news listing.
 
@@ -457,6 +514,7 @@ class Client(Eventable):
             resp.next = partial(self.get_news_listing, **kwargs)
         return resp
 
+    @prepare_client
     async def get_news_post(
         self, news_query: Union[str, int], **kwargs: Any
     ) -> NewsPost:
@@ -484,6 +542,7 @@ class Client(Eventable):
         json = await self._request("GET", url, params=params)
         return NewsPost.parse_obj(json)
 
+    @prepare_client
     async def get_wiki_page(self, locale: str, path: str) -> WikiPage:
         r"""Gets a wiki page.
 
@@ -499,6 +558,7 @@ class Client(Eventable):
         json = await self._request("GET", url)
         return WikiPage.parse_obj(json)
 
+    @prepare_client
     async def get_comment(self, comment_id: int, **kwargs: Any) -> CommentBundle:
         r"""Gets a comment.
 
@@ -525,6 +585,7 @@ class Client(Eventable):
             resp.next = partial(self.get_comment, comment_id=comment_id, **kwargs)
         return resp
 
+    @prepare_client
     async def get_comments(self, **kwargs: Any) -> CommentBundle:
         r"""Gets comments.
 
@@ -561,6 +622,7 @@ class Client(Eventable):
             resp.next = partial(self.get_comments, **kwargs)
         return resp
 
+    @prepare_client
     @check_token
     async def search(self, query: str, **kwargs: Any) -> SearchResponse:
         r"""Searches for a user, beatmap, beatmapset, or wiki page.
@@ -589,6 +651,7 @@ class Client(Eventable):
         json = await self._request("GET", url, params=params)
         return SearchResponse.parse_obj(json)
 
+    @prepare_client
     @check_token
     @requires_scope(Scopes.IDENTIFY)
     async def get_me(self, **kwargs: Any) -> User:
@@ -612,6 +675,7 @@ class Client(Eventable):
         json = await self._request("GET", url)
         return User.parse_obj(json)
 
+    @prepare_client
     @check_token
     @requires_scope(Scopes.FRIENDS_READ)
     async def get_own_friends(self) -> list[User]:
@@ -625,6 +689,7 @@ class Client(Eventable):
         json = await self._request("GET", url)
         return from_list(User.parse_obj, json)
 
+    @prepare_client
     @check_token
     async def get_user(self, user_query: Union[str, int], **kwargs: Any) -> User:
         r"""Gets a user by a query.
@@ -659,6 +724,7 @@ class Client(Eventable):
         json = await self._request("GET", url, params=params)
         return User.parse_obj(json)
 
+    @prepare_client
     @check_token
     async def get_users(self, user_ids: list[int]) -> list[User]:
         r"""Get multiple user data.
@@ -676,6 +742,7 @@ class Client(Eventable):
         json = await self._request("GET", url, params=params)
         return from_list(User.parse_obj, json.get("users", []))
 
+    @prepare_client
     @check_token
     async def get_user_kudosu(self, user_id: int, **kwargs: Any) -> list[KudosuHistory]:
         r"""Get a user's kudosu history.
@@ -702,6 +769,7 @@ class Client(Eventable):
         json = await self._request("GET", url, params=params)
         return from_list(KudosuHistory.parse_obj, json)
 
+    @prepare_client
     @check_token
     async def __get_type_scores(
         self, user_id: int, request_type: str, **kwargs: Any
@@ -819,6 +887,7 @@ class Client(Eventable):
         """
         return await self.__get_type_scores(user_id, "firsts", **kwargs)
 
+    @prepare_client
     @check_token
     async def get_user_beatmap_scores(
         self, user_id: int, beatmap_id: int, **kwargs: Any
@@ -848,6 +917,7 @@ class Client(Eventable):
 
     UserBeatmapType = Literal["favourite", "graveyard", "loved", "ranked", "pending"]
 
+    @prepare_client
     @check_token
     async def get_user_beatmaps(
         self, user_id: int, type: UserBeatmapType, **kwargs: Any
@@ -878,6 +948,7 @@ class Client(Eventable):
         json = await self._request("GET", url, params=params)
         return from_list(Beatmapset.parse_obj, json)
 
+    @prepare_client
     @check_token
     async def get_user_most_played(
         self, user_id: int, **kwargs: Any
@@ -906,6 +977,7 @@ class Client(Eventable):
         json = await self._request("GET", url, params=params)
         return from_list(BeatmapUserPlaycount.parse_obj, json)
 
+    @prepare_client
     @check_token
     async def get_user_recent_activity(
         self, user_id: int, **kwargs: Any
@@ -934,6 +1006,7 @@ class Client(Eventable):
         json = await self._request("GET", url, params=params)
         return from_list(Event.parse_obj, json)
 
+    @prepare_client
     @check_token
     async def get_beatmap_scores(self, beatmap_id: int, **kwargs: Any) -> list[Score]:
         r"""Get scores submitted on a specific beatmap.
@@ -963,6 +1036,7 @@ class Client(Eventable):
         json = await self._request("GET", url, params=params)
         return from_list(Score.parse_obj, json.get("scores", []))
 
+    @prepare_client
     @check_token
     async def get_beatmap(self, beatmap_id: int) -> Beatmap:
         r"""Get beatmap data.
@@ -977,6 +1051,7 @@ class Client(Eventable):
         json = await self._request("GET", url)
         return Beatmap.parse_obj(json)
 
+    @prepare_client
     @check_token
     async def get_beatmaps(self, beatmap_ids: list[int]) -> list[Beatmap]:
         r"""Get multiple beatmap data.
@@ -994,6 +1069,7 @@ class Client(Eventable):
         json = await self._request("GET", url, params=params)
         return from_list(Beatmap.parse_obj, json.get("beatmaps", []))
 
+    @prepare_client
     @check_token
     async def lookup_beatmap(self, **kwargs: Any) -> Beatmap:
         r"""Lookup beatmap data.
@@ -1024,6 +1100,7 @@ class Client(Eventable):
         json = await self._request("GET", url, params=params)
         return Beatmap.parse_obj(json)
 
+    @prepare_client
     @check_token
     async def get_beatmap_attributes(
         self, beatmap_id: int, **kwargs: Any
@@ -1058,6 +1135,7 @@ class Client(Eventable):
         json = await self._request("POST", url, data=data)
         return BeatmapDifficultyAttributes.parse_obj(json.get("attributes"))
 
+    @prepare_client
     @check_token
     async def get_beatmapset(self, beatmapset_id: int) -> Beatmapset:
         r"""Get beatmapset data.
@@ -1072,6 +1150,7 @@ class Client(Eventable):
         json = await self._request("GET", url)
         return Beatmapset.parse_obj(json)
 
+    @prepare_client
     @check_token
     async def lookup_beatmapset(self, beatmap_id: int) -> Beatmapset:
         r"""Lookup beatmap data.
@@ -1090,6 +1169,7 @@ class Client(Eventable):
         json = await self._request("GET", url, params=params)
         return Beatmapset.parse_obj(json)
 
+    @prepare_client
     @check_token
     async def search_beatmapsets(
         self,
@@ -1121,6 +1201,7 @@ class Client(Eventable):
             resp.next = partial(self.search_beatmapsets, **kwargs)
         return resp
 
+    @prepare_client
     @check_token
     async def get_beatmapset_events(self, **kwargs: Any) -> list[BeatmapsetEvent]:
         r"""Get beatmapset events.
@@ -1157,6 +1238,7 @@ class Client(Eventable):
         json = await self._request("GET", url, params=params)
         return from_list(BeatmapsetEvent.parse_obj, json.get("events", []))
 
+    @prepare_client
     @check_token
     async def get_beatmapset_discussions(
         self, **kwargs: Any
@@ -1214,6 +1296,7 @@ class Client(Eventable):
             resp.next = partial(self.get_beatmapset_discussions, **kwargs)
         return resp
 
+    @prepare_client
     @check_token
     async def get_beatmapset_discussion_posts(
         self, **kwargs: Any
@@ -1262,6 +1345,7 @@ class Client(Eventable):
             resp.next = partial(self.get_beatmapset_discussion_posts, **kwargs)
         return resp
 
+    @prepare_client
     @check_token
     async def get_beatmapset_discussion_votes(
         self, **kwargs: Any
@@ -1313,6 +1397,7 @@ class Client(Eventable):
             resp.next = partial(self.get_beatmapset_discussion_votes, **kwargs)
         return resp
 
+    @prepare_client
     @check_token
     async def get_score(
         self,
@@ -1334,6 +1419,7 @@ class Client(Eventable):
         json = await self._request("GET", url)
         return Score.parse_obj(json)
 
+    @prepare_client
     @check_token
     @requires_scope(Scopes.IDENTIFY | Scopes.DELEGATE, any_scope=True)
     async def get_score_replay(
@@ -1355,6 +1441,7 @@ class Client(Eventable):
         url = f"{self.base_url}/api/v2/scores/{mode}/{score_id}/download"
         return await self._request("GET", url)
 
+    @prepare_client
     @check_token
     async def get_rankings(
         self, mode: Gamemode, type: RankingType, **kwargs: Any
@@ -1398,6 +1485,7 @@ class Client(Eventable):
             resp.next = partial(self.get_rankings, mode=mode, type=type, **kwargs)
         return resp
 
+    @prepare_client
     @check_token
     async def get_spotlights(self) -> list[Spotlight]:
         r"""Gets the current spotlights.
@@ -1410,6 +1498,7 @@ class Client(Eventable):
         json = await self._request("GET", url)
         return from_list(Spotlight.parse_obj, json.get("spotlights", []))
 
+    @prepare_client
     @check_token
     async def get_forum_topic(self, topic_id: int, **kwargs: Any) -> ForumTopicResponse:
         r"""Gets a forum topic.
@@ -1452,6 +1541,7 @@ class Client(Eventable):
             resp.next = partial(self.get_forum_topic, topic_id, **kwargs)
         return resp
 
+    @prepare_client
     @check_token
     @requires_scope(Scopes.FORUM_WRITE)
     async def create_forum_topic(
@@ -1513,6 +1603,7 @@ class Client(Eventable):
         json = await self._request("POST", url, data=data)
         return ForumCreateTopicResponse.parse_obj(json)
 
+    @prepare_client
     @check_token
     @requires_scope(Scopes.FORUM_WRITE)
     async def reply_forum_topic(self, topic_id: int, content: str) -> ForumPost:
@@ -1533,6 +1624,7 @@ class Client(Eventable):
         json = await self._request("POST", url, data=data)
         return ForumPost.parse_obj(json)
 
+    @prepare_client
     @check_token
     @requires_scope(Scopes.FORUM_WRITE)
     async def edit_forum_topic_title(self, topid_id: int, new_title: str) -> ForumTopic:
@@ -1555,6 +1647,7 @@ class Client(Eventable):
         json = await self._request("PUT", url, data=data)
         return ForumTopic.parse_obj(json)
 
+    @prepare_client
     @check_token
     @requires_scope(Scopes.FORUM_WRITE)
     async def edit_forum_post(self, post_id: int, new_content: str) -> ForumPost:
@@ -1575,6 +1668,7 @@ class Client(Eventable):
         json = await self._request("PUT", url, data=data)
         return ForumPost.parse_obj(json)
 
+    @prepare_client
     @check_token
     @requires_scope(Scopes.LAZER)
     async def get_chat_ack(self, **kwargs: Any) -> list[ChatUserSilence]:
@@ -1600,6 +1694,7 @@ class Client(Eventable):
         json = await self._request("POST", url, data=data)
         return from_list(ChatUserSilence.parse_obj, json.get("silences", []))
 
+    @prepare_client
     @check_token
     @requires_scope(Scopes.LAZER)
     async def get_chat_updates(self, since: int, **kwargs: Any) -> ChatUpdateResponse:
@@ -1638,6 +1733,7 @@ class Client(Eventable):
         json = await self._request("GET", url, data=data)
         return ChatUpdateResponse.parse_obj(json)
 
+    @prepare_client
     @check_token
     @requires_scope(Scopes.LAZER)
     async def get_channel(self, channel_id: int) -> ChatChannelResponse:
@@ -1653,6 +1749,7 @@ class Client(Eventable):
         json = await self._request("GET", url)
         return ChatChannelResponse.parse_obj(json)
 
+    @prepare_client
     @check_token
     @requires_scope(Scopes.LAZER)
     async def get_channels(self) -> list[ChatChannel]:
@@ -1666,6 +1763,7 @@ class Client(Eventable):
         json = await self._request("GET", url)
         return from_list(ChatChannel.parse_obj, json)
 
+    @prepare_client
     @check_token
     @requires_scope(Scopes.LAZER)
     async def get_channel_messages(
@@ -1702,6 +1800,7 @@ class Client(Eventable):
         json = await self._request("GET", url, data=data)
         return from_list(ChatMessage.parse_obj, json)
 
+    @prepare_client
     @check_token
     @requires_scope(Scopes.CHAT_WRITE)
     async def create_chat_channel(
@@ -1715,17 +1814,18 @@ class Client(Eventable):
             See below
 
         :Keyword Arguments:
-            * *target_id* (``int``) --
-                Required if type is ``PM``, the ID of the user to send a PM to
-            * *target_ids* (``List[int]``) --
-                Required if type is ``PM``, the IDs of the users to send a PM to
             * *message* (``str``) --
                 Required if type is ``ANNOUNCE``, the message to send in the PM
+            * *target_id* (``int``) --
+                Only used if if type is ``PM``, the ID of the user to send a PM to
+            * *target_ids* (``List[int]``) --
+                Only used if type is ``ANNOUNCE``, the IDs of the users to send a PM to
             * *channel_name* (``str``) --
-                Required if type is ``ANNOUNCE``, the name of the channel
+                Only used if type is ``ANNOUNCE``, the name of the channel
             * *channel_description* (``str``) --
-                Required if type is ``ANNOUNCE``, the description of the channel
+                Only used if type is ``ANNOUNCE``, the description of the channel
 
+        :raises ValueError: If missing required parameters
         :raises APIException: Contains status code and error message
         :return: Chat channel object
         :rtype: aiosu.models.chat.ChatChannel
@@ -1736,11 +1836,11 @@ class Client(Eventable):
         }
         add_param(data, kwargs, key="message")
         if type == "PM":
-            has_id = add_param(data, kwargs, key="target_id")
-            has_id |= add_param(data, kwargs, key="target_ids")
-            if not has_id:
-                raise ValueError("Missing target ID(s)")
+            if not add_param(data, kwargs, key="target_id"):
+                raise ValueError("Missing target ID")
         elif type == "ANNOUNCE":
+            if not add_param(data, kwargs, key="target_ids"):
+                raise ValueError("Missing target IDs")
             if not data.get("message"):
                 raise ValueError("Missing message")
             channel = {
@@ -1751,6 +1851,7 @@ class Client(Eventable):
         json = await self._request("POST", url, data=data)
         return ChatChannel.parse_obj(json)
 
+    @prepare_client
     @check_token
     @requires_scope(Scopes.LAZER)
     async def join_channel(self, channel_id: int, user_id: int) -> ChatChannel:
@@ -1768,6 +1869,7 @@ class Client(Eventable):
         json = await self._request("PUT", url)
         return ChatChannel.parse_obj(json)
 
+    @prepare_client
     @check_token
     @requires_scope(Scopes.LAZER)
     async def leave_channel(self, channel_id: int, user_id: int) -> None:
@@ -1782,6 +1884,7 @@ class Client(Eventable):
         url = f"{self.base_url}/api/v2/chat/channels/{channel_id}/users/{user_id}"
         await self._request("DELETE", url)
 
+    @prepare_client
     @check_token
     @requires_scope(Scopes.LAZER)
     async def mark_read(self, channel_id: int, message_id: int) -> None:
@@ -1796,6 +1899,7 @@ class Client(Eventable):
         url = f"{self.base_url}/api/v2/chat/channels/{channel_id}/mark-as-read/{message_id}"
         await self._request("PUT", url)
 
+    @prepare_client
     @check_token
     @requires_scope(Scopes.LAZER)
     async def send_message(
@@ -1824,6 +1928,7 @@ class Client(Eventable):
         json = await self._request("POST", url, data=data)
         return ChatMessage.parse_obj(json)
 
+    @prepare_client
     @check_token
     @requires_scope(Scopes.CHAT_WRITE)
     async def send_private_message(
@@ -1842,7 +1947,7 @@ class Client(Eventable):
 
         :Keyword Arguments:
             * *uuid* (``str``) --
-                Optional, the UUID of the message
+                Optional, client-side message identifier to be sent back in response and websocket json
 
         :raises APIException: Contains status code and error message
         :return: Chat message create response object
@@ -1858,6 +1963,7 @@ class Client(Eventable):
         json = await self._request("POST", url, data=data)
         return ChatMessageCreateResponse.parse_obj(json)
 
+    @prepare_client
     @check_token
     async def get_multiplayer_matches(
         self, **kwargs: Any
@@ -1894,6 +2000,7 @@ class Client(Eventable):
             resp.next = partial(self.get_multiplayer_matches, **kwargs)
         return resp
 
+    @prepare_client
     @check_token
     async def get_multiplayer_match(
         self, match_id: int, **kwargs: Any
@@ -1929,6 +2036,7 @@ class Client(Eventable):
         json = await self._request("GET", url)
         return MultiplayerMatchResponse.parse_obj(json)
 
+    @prepare_client
     @check_token
     async def get_multiplayer_rooms(self, **kwargs: Any) -> list[MultiplayerRoom]:
         r"""Gets the multiplayer rooms.
@@ -1968,6 +2076,7 @@ class Client(Eventable):
         json = await self._request("GET", url, params=params)
         return from_list(MultiplayerRoom.parse_obj, json)
 
+    @prepare_client
     @check_token
     async def get_multiplayer_room(self, room_id: int) -> MultiplayerRoom:
         r"""Gets a multiplayer room.
@@ -1983,6 +2092,7 @@ class Client(Eventable):
         json = await self._request("GET", url)
         return MultiplayerRoom.parse_obj(json)
 
+    @prepare_client
     @check_token
     async def get_multiplayer_leaderboard(
         self, room_id: int, **kwargs: Any
@@ -2012,6 +2122,7 @@ class Client(Eventable):
         json = await self._request("GET", url, params=params)
         return MultiplayerLeaderboardResponse.parse_obj(json)
 
+    @prepare_client
     @check_token
     @requires_scope(Scopes.IDENTIFY | Scopes.DELEGATE, any_scope=True)
     async def get_multiplayer_scores(
@@ -2056,6 +2167,7 @@ class Client(Eventable):
             )
         return resp
 
+    @prepare_client
     @check_token
     async def revoke_token(self) -> None:
         r"""Revokes the current token and closes the session.
